import warnings
from collections import deque
from typing import Optional

import numpy as np
from matplotlib.collections import LineCollection
from mpl_toolkits.axes_grid1 import make_axes_locatable

from dustyn.core.transform import spherical2cartesian

NAME2COL = {"x": 0, "y": 1, "z": 2}


class Projected3DTrajectoryPlot:
    """
    Plot a 3D trajectory, a represented in a Cartesian coordinate system,
    in 2 dimension. The normal coordinate is shown in color.
    """

    def __init__(
        self,
        ax,
        fig,
        y,
        normal: str = "z",
        input_geometry="cartesian",
        **kwargs,
    ):

        if input_geometry == "spherical":
            y = spherical2cartesian(position=y).T
        elif input_geometry == "cartesian":
            pass
        else:
            raise ValueError(f"Unrecognized value `{input_geometry=}`.")

        axes_names = deque("xyz")
        if normal not in axes_names:
            raise ValueError

        while not axes_names[-1] == normal:
            axes_names.rotate()

        labels = list(axes_names)[:2]
        ax.set(xlabel=r"$%s$" % labels[0], ylabel=r"$%s$" % labels[1], aspect="equal")
        x1, x2 = (y[:, NAME2COL[v]] for v in labels)

        self.setup_plot(ax, fig, y, x1, x2, normal, **kwargs)

    def setup_plot(
        self,
        ax,
        fig,
        y,
        x1,
        x2,
        normal: str,
        colors: Optional[np.ndarray] = None,
        zlabel: Optional[str] = None,
        **kwargs,
    ):
        if colors is None:
            x3 = y[:, NAME2COL[normal]]
            if zlabel is not None:
                warnings.warn(
                    "Received `zlabel` but `colors` keyword argument is unspecified, "
                    "`zlabel` will be ignored."
                )
            zlabel = normal
        else:
            x3 = colors

        points = np.array([x1, x2]).T.reshape(-1, 1, 2)

        # can probably be done in a cleaner way
        segments = np.concatenate([points[:-1], points[1:]], axis=1)

        # this is adapted from
        # https://matplotlib.org/stable/gallery/lines_bars_and_markers/multicolored_line.html

        # Create a continuous norm to map from data points to colors
        lc = LineCollection(segments, **kwargs)
        # Set the values used for colormapping
        lc.set_array(x3)
        lc.set_linewidth(kwargs.get("lw", 2))
        line = ax.add_collection(lc)

        # this part controls that the colorbar fits the size of the plot
        # adapted from
        # https://stackoverflow.com/questions/29516157/set-equal-aspect-in-plot-with-colorbar
        divider = make_axes_locatable(ax)
        color_axis = divider.append_axes("right", size="5%", pad=0)

        cbar_kwargs = dict(cax=color_axis, pad=0)
        if zlabel is not None:
            # this is a workaround a bug in matplotlib
            cbar_kwargs.update(dict(label=r"$%s$" % zlabel))
        fig.colorbar(line, **cbar_kwargs)


class SimpleProjected3DTrajectoryPlot(Projected3DTrajectoryPlot):
    def setup_plot(self, ax, fig, y, x1, x2, normal, **kwargs):
        ax.plot(x1, x2, **kwargs)



import numpy as np


def cross_product(a: np.ndarray, b: np.ndarray, /) -> np.ndarray:
    """
    This implementation is 10x faster than np.cross in my applications.
    """
    return np.array(
        [
            a[1] * b[2] - a[2] * b[1],
            a[2] * b[0] - a[0] * b[2],
            a[0] * b[1] - a[1] * b[0],
        ]
    )



from typing import Optional
from typing import Sequence

import numpy as np
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import ProgressColumn
from rich.progress import SpinnerColumn
from rich.progress import Task
from rich.progress import Text
from rich.progress import TextColumn
from rich.progress import TimeElapsedColumn
from rich.progress import TimeRemainingColumn

from dustyn._typing import FloatLike


class CompletionColumn(ProgressColumn):
    """Renders the current number of iterations completed."""

    def render(self, task: Task) -> Text:
        return Text(f"{task.completed} it", style="progress.percentage")


class CompletionSpeedColumn(ProgressColumn):
    """Renders human readable completion speed."""

    # This is based off rich.progress.TransferSpeedColumn

    def render(self, task: Task) -> Text:
        """Show data transfer speed."""
        speed = task.finished_speed or task.speed
        if speed is None:
            return Text("?", style="progress.data.speed")
        data_speed = int(speed)
        return Text(f"{data_speed} it/s", style="progress.data.speed")


class RichProgressBar:
    """A thin wrapper around rich.progress.Progress"""

    def __init__(
        self,
        title: str,
        total: Optional[FloatLike] = None,
        columns: Optional[Sequence[ProgressColumn]] = None,
    ):
        if columns is None:
            columns = [
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
            ]
            if total is None:
                columns += [
                    CompletionColumn(),
                    TextColumn("[progress.data.speed]@"),
                    CompletionSpeedColumn(),
                ]
                # this is a workaround. As of rich 9.13.0, some of the columns
                # we're using here require a non-None value in `task.total` to render
                total = np.inf
            else:
                columns += [
                    BarColumn(),
                    TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                    TimeRemainingColumn(),
                ]
            columns += [TimeElapsedColumn()]
        self._pbar = Progress(*columns)

        self._pbar.start()
        self.task = self._pbar.add_task(title, total=total)

    def update(
        self,
        *,
        completed: Optional[FloatLike] = None,
        advance: Optional[FloatLike] = None,
        **kwargs,
    ) -> None:
        # We only expose the relevant subset of keyword arguments
        # supported by `rich.progress.Progress.update`
        # note that they should NOT be combined (use either `completed` or `advance` but not both)
        self._pbar.update(self.task, completed=completed, advance=advance, **kwargs)

    def refresh(self):
        self._pbar.refresh()

    def stop(self):
        self._pbar.stop()


def get_pbar(
    title: str, total: Optional[FloatLike] = None, start: bool = True
) -> Progress:
    return RichProgressBar(title, total)


class DummyProgressBar:
    # mock rich.progress.Progress's interface
    def update(
        self,
        *,
        completed: Optional[FloatLike] = None,
        advance: Optional[FloatLike] = None,
        **kwargs,
    ) -> None:
        pass

    def refresh(self):
        pass

    def stop(self):
        pass


if __name__ == "__main__":
    from time import sleep

    pbar = get_pbar(title="progress with advance...", total=70.0)
    for _ in range(70):
        sleep(0.02)
        pbar.update(advance=1)
    pbar.stop()

    pbar = get_pbar(title="progress with completed...", total=70.0)
    for i in range(70):
        sleep(0.02)
        pbar.update(completed=i + 1)
    pbar.stop()



from typing import Any
from typing import Callable
from typing import Optional

import numpy as np
from numpy import cos
from numpy import ScalarType
from numpy import sin


def _get_rotation_matrix(r: ScalarType, theta: ScalarType, phi: ScalarType):
    return np.array(
        [
            [sin(theta) * cos(phi), sin(theta) * sin(phi), cos(theta)],
            [cos(theta) * cos(phi), cos(theta) * sin(phi), -sin(theta)],
            [-sin(phi), cos(phi), 0],
        ],
    )


def spherical2cartesian(
    position: np.ndarray, vector: Optional[np.ndarray] = None
) -> np.ndarray:
    """
    Transform a spherical `position` (r, theta, phi) coordinate array
    to cartesian (x, y, z).

    When `vector` is provided, transform this instead.
    """
    assert 3 in position.shape
    if len(position) != 3:
        position = position.T
    r, theta, phi = position

    if isinstance(r, ScalarType):
        MROT = _get_rotation_matrix(*position)
        if vector is None:
            return [r, 0, 0] @ MROT
        return vector @ MROT

    assert r.ndim == 1

    ret = np.empty_like(position, order="F")
    for i in range(len(r)):
        if vector is not None:
            v = vector[:, i]
        else:
            v = None
        ret[:, i] = spherical2cartesian(position=position[:, i], vector=v)
    return ret


def spherical2cylindrical(position: np.ndarray) -> np.ndarray:
    assert position.ndim == 2
    assert 3 in position.shape
    if len(position) != 3:
        position = position.T
    r, theta, phi = position
    return np.array([r * np.sin(theta), phi, r * np.cos(theta)])


def cylindrical2spherical(position: np.ndarray) -> np.ndarray:
    assert 3 in position.shape
    if len(position) != 3:
        position = position.T
    R, phi, z = position
    return np.array([np.sqrt(R ** 2 + z ** 2), np.arctan2(R, z), phi])


# this is meant to be imported from outside
TRANSFORMATIONS: dict[str, dict[str, Callable[[Any], np.ndarray]]] = {
    "spherical": {
        "cartesian": spherical2cartesian,
        "cylindrical": spherical2cylindrical,
    },
    "cylindrical": {
        "spherical": cylindrical2spherical,
    },
}



from typing import Optional
from typing import Sequence

import numpy as np
from matplotlib.axes import Axes

from dustyn._typing import SingleOrDouble


class SpaceSampler:

    """Create a 2D discrete rectilinear grid specified by bounds, geometry and spacing type."""

    _instance = None

    def __new__(cls):
        # Implement the singleton pattern
        # Long term, this should be migrated to a global object
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @staticmethod
    def get_grid(
        *,
        bounds: np.ndarray,
        npoints: tuple[int, int] = (100, 100),
        spacing: tuple[str, str] = ("linear", "linear"),
    ) -> np.ndarray:

        if len(bounds) != 4:
            raise ValueError(f"Expected a 4-size `bounds` array, received `{bounds}`.")

        if spacing != ("linear", "linear"):
            raise NotImplementedError

        if not isinstance(spacing, Sequence):
            raise TypeError(f"Expected a sequence, received `{type(spacing)}`.")

        if spacing == ("linear", "linear"):
            space_discretization = np.linspace

        xv = space_discretization(*bounds[0:2], npoints[0])
        yv = space_discretization(*bounds[2:4], npoints[1])

        return np.meshgrid(xv, yv)

    def __call__(
        self,
        *,
        bounds: Optional[np.ndarray] = None,
        npoints: SingleOrDouble[int] = 100,
        ax: Optional[Axes] = None,
        geometry: str = "cartesian",
        spacing: SingleOrDouble[str] = "linear",
    ) -> np.ndarray:

        if geometry != "cartesian":
            raise NotImplementedError

        if bounds is None:
            if geometry != "cartesian":
                raise TypeError(
                    "`bounds` keyword argument is required with `geometry` != 'cartesian'."
                )
            if ax is None:
                raise TypeError(
                    "Either `bounds` or `ax` keyword argument must be specified."
                )

            bounds = [*ax.get_xlim(), *ax.get_ylim()]

        if isinstance(npoints, int):
            npoints = (npoints, npoints)
        if isinstance(spacing, str):
            spacing = (spacing, spacing)
        return self.get_grid(bounds=bounds, npoints=npoints, spacing=spacing)



import weakref
from collections import deque
from dataclasses import dataclass
from typing import Optional
from typing import Union

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes._axes import Axes
from matplotlib.collections import LineCollection
from mpl_toolkits.axes_grid1 import make_axes_locatable

from dustyn.core.record import Record
from dustyn.core.transform import TRANSFORMATIONS

CURVILINEAR_AXES = frozenset({"phi", "theta"})
AXES: dict[str, list[str]] = {
    "cartesian": ["x", "y", "z"],
    "cylindrical": ["R", "phi", "z"],
    "spherical": ["r", "theta", "phi"],
}
LATEX_SYMBOLS: dict[str, str] = {
    "phi": r"$\varphi$",
    "theta": r"$\theta$",
}


def texify_symbol(s: str) -> str:
    return LATEX_SYMBOLS.get(s, s)


@dataclass
class Vector:
    data: np.ndarray
    label: str

    def __len__(self):
        # NOTE: this could be abstracted into a DataContainer abstract dataclass
        # since it's common to the Record class
        return len(self.data)

    @property
    def shape(self):
        return self.data.shape


class RecordPlot:
    def __init__(
        self,
        record: Record,
        ax: Optional[Axes] = None,
        *,
        geometry: str = "cartesian",
        normal: Optional[str] = None,
        clabel: Optional[str] = None,
        colors: Optional[Union[np.ndarray, str]] = None,
        cbar: bool = True,
        logscale: bool = False,
        L0=1.0,
        **kwargs,
    ):
        if ax is None:
            _, ax = plt.subplots()
            self.ax = ax
        else:
            self.ax = weakref.proxy(ax)
        self.record = weakref.proxy(record)
        self.geometry = geometry
        self.axes_labels = AXES[self.geometry]
        normal = normal or AXES[self.geometry][2]
        self._flip_axes = normal.startswith("-")
        if self._flip_axes:
            normal = normal[1:]
        self.normal = normal
        self.L0 = L0

        self._setup_transform()

        x1, x2 = self._get_plane_vectors()

        x3: Optional[Vector]
        if isinstance(colors, str):
            if colors == "normal":
                x3 = self._get_normal_vector()
            elif colors == "time":
                x3 = Vector(data=record.times, label=r"$t$")
            else:
                raise ValueError
        elif isinstance(colors, np.ndarray):
            x3 = Vector(data=colors, label=clabel or "")
        elif colors is None:
            x3 = None
        else:
            raise TypeError

        self._setup_plot(x1, x2, x3, **kwargs)

        if logscale:
            if x1.label not in CURVILINEAR_AXES:
                ax.set_xscale("log")
            if x2.label not in CURVILINEAR_AXES:
                ax.set_yscale("log")

        if cbar and colors is not None:
            self.setup_cbar()

    def _setup_transform(self) -> None:
        record_geometry = self.record.metadata["geometry"]
        if self.geometry == record_geometry:
            self.states = self.record.states[:, 3:].T
        else:
            try:
                transform = TRANSFORMATIONS[record_geometry][self.geometry]
            except KeyError as err:
                raise NotImplementedError from err
            self.states = transform(self.record.states[:, 3:])

    def normalize_data(self, axis_name: str) -> Vector:
        data = self.states[self.axes_labels.index(axis_name)]
        if axis_name in CURVILINEAR_AXES:
            data %= 2 * np.pi
        else:
            data *= self.L0
        return Vector(data, label=axis_name)

    def _get_plane_vectors(self) -> tuple[Vector, Vector]:
        names = deque(self.axes_labels)
        while names[-1] != self.normal:
            names.rotate()
        labels = list(names)[:2]
        if self._flip_axes:
            labels = labels[::-1]
        # deactivating mypy here because it seems to be insecure about the return
        # tuple length, but it should be obvious from the code above that
        # it will always be 2 and match the signature
        return tuple(self.normalize_data(label) for label in labels)  # type: ignore

    def _get_normal_vector(self) -> Vector:
        if not (label := self.normal) in self.axes_labels:
            raise ValueError(
                f"Unknown normal axis '{self.normal}' with geometry '{self.geometry}'."
            )
        return self.normalize_data(label)

    def set(self, **kwargs):
        self.ax.set(**kwargs)

    def _setup_plot(
        self,
        x1: Vector,
        x2: Vector,
        x3: Optional[Vector],
        **kwargs,
    ) -> None:

        self.set(xlabel=texify_symbol(x1.label), ylabel=texify_symbol(x2.label))
        if not any(label in CURVILINEAR_AXES for label in (x1.label, x2.label)):
            self.set(aspect="equal")

        if x3 is None:
            self.ax.plot(x1.data, x2.data, **kwargs)
            return

        # this is a workaround to compensate for the fact that
        # LineCollection doesn't force xlim and ylim to auto adjust,
        # so we'll draw a basic "plot" in even if using colors,
        # but we'll make it invisible
        self.ax.plot(x1.data, x2.data, lw=0)

        points = np.array([x1.data, x2.data]).T.reshape(-1, 1, 2)

        # can probably be done in a cleaner way
        segments = np.concatenate([points[:-1], points[1:]], axis=1)

        # this is adapted from
        # https://matplotlib.org/stable/gallery/lines_bars_and_markers/multicolored_line.html

        # Create a continuous norm to map from data points to colors
        lc = LineCollection(segments, **kwargs)
        # Set the values used for colormapping
        lc.set_array(x3.data)
        lc.set_linewidth(kwargs.get("lw", 2))
        self._mappable = self.ax.add_collection(lc)
        self.x3 = x3

    def setup_cbar(self):
        # Controls that the colorbar fits the size of the plot

        # NOTE: I can't seem to find one method that works well in every situation (log/lin).
        # Additional difficulties arise when the size of the figure is updated after the cbar
        # was setup, so the recommended way to use this is to run any call to
        # ax.set_xlim, ax.set_ylim, fig.subplots_adjust...
        # ahead of plotting, and *then* call this method. This is why it is separated from the
        # _setup_plot method, and also why it is written as user-facing.

        if self.ax.get_yscale() == "log":
            axpos = self.ax.get_position()
            # args are:
            # left, bottom, width, height
            cax = self.ax.figure.add_axes(
                [axpos.x1, axpos.y0, (axpos.x1 - axpos.x0) * 0.05, axpos.height]
            )
        else:
            # this works extremely poorly with logscale
            # it is adapted from
            # https://stackoverflow.com/questions/29516157/set-equal-aspect-in-plot-with-colorbar
            divider = make_axes_locatable(self.ax)
            cax = divider.append_axes("right", size="5%", pad=0)

        self.ax.figure.colorbar(
            self._mappable, cax=cax, label=texify_symbol(self.x3.label)
        )



import abc
import sys
import warnings
from datetime import datetime
from time import ctime
from typing import Callable
from typing import Optional

import numpy as np

from .evolve import EvolutionModel
from .progress import DummyProgressBar
from .progress import get_pbar
from dustyn._typing import FloatLike
from dustyn.core.record import Record


class SolverRecord(Record):
    def __init__(self, *args, **kwargs):

        print(
            "warning: the SolverRecord class is now identical to Record (except for this warning).",
            file=sys.stderr,
        )
        super().__init__(*args, **kwargs)


class Solver(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def step(
        R: np.ndarray, time: FloatLike, timestep: FloatLike, func: Callable
    ) -> None:
        """This should walk the R vector state over one timestep according to the evolution
        function `func` with signature

        def func(time, R):
            ...
        """
        pass

    @staticmethod
    def _update_from_buffer(output: list[np.ndarray], buffer=np.ndarray) -> None:
        output.append(buffer.copy())

    @classmethod
    def integrate(
        cls,
        model: EvolutionModel,
        R0: np.ndarray,
        *,
        tstart: FloatLike,
        tstop: FloatLike,
        CFL: FloatLike,
        pbar: bool = False,
        buffer_size: int = 1000,
        dt_record: Optional[FloatLike] = None,
        dt_min: FloatLike = 0.0,
        label: Optional[str] = None,
        logging: bool = False,
    ) -> Record:

        if CFL > 1.0:
            warnings.warn(
                f"Received {CFL=} > 1, stability is not guaranteed and results will likely be inaccurate."
            )
        elif CFL < 0:
            raise ValueError("Received negative CFL security.")
        elif CFL < (min_float := np.finfo(R0.dtype).eps):
            raise ValueError(
                f"Received too low value for {CFL=} . The minimal value is {min_float}"
            )

        tstart, tstop, CFL = map(R0.dtype.type, (tstart, tstop, CFL))

        if pbar:
            if buffer_size < 1000:
                warnings.warn(
                    "refreshing the progress bar too often is extremely expensive. "
                    "It is strongly advised to use progress bars only with `buffer_size` >= 1000."
                )
            if (
                dt_record is not None
                and int((tstop - tstart) // dt_record) <= 2 * buffer_size
            ):
                warnings.warn(
                    "Time resolution is too low to allow useful granulosity in the progressbar."
                )
            if logging:
                warnings.warn(
                    "logging and pbar can not be used simultaneously. Turning logging off."
                )
                logging = False

            title = label if label is not None else "Integrating..."
            progress = get_pbar(title=title, total=tstop)
        else:
            progress = DummyProgressBar()

        metadata = dict(
            CFL=CFL,
            label=label,
            geometry=getattr(model, "geometry", None),
            dimensionality=getattr(model, "dimensionality", None),
            stop_reason="completed",
        )
        states_buffer = np.empty((buffer_size, *R0.shape), dtype=R0.dtype)
        times_buffer = np.empty(buffer_size, dtype=R0.dtype)

        buffer_pos = 0
        R = R0.copy()
        t = tstart

        times: list[np.ndarray] = [np.array([tstart])]
        states: list[np.ndarray] = [np.expand_dims(R0, axis=0)]
        rec_count = 1
        next_trec = dt_record
        timestep_phy = CFL * model.get_max_timestep(R)
        it = 0
        while t < tstop:
            if dt_record is None:
                timestep_rec = min(timestep_phy, tstop - t)
                next_trec = t + timestep_rec
            else:
                timestep_rec = next_trec - t
            to_rec: bool = timestep_rec <= timestep_phy
            timestep = timestep_rec if to_rec else timestep_phy

            cls.step(R, time=t, timestep=timestep, func=model.evolve)

            t += timestep

            # this is essentially equivalent to
            # >>> timestep = min(CFL * model.get_max_timestep(R), tsave - t)
            # but with the advantage of being floating-point comparison proof
            # when we need to decide wether or not to save later.
            timestep_phy = CFL * model.get_max_timestep(R)
            nan_timestep: bool = timestep_phy != timestep_phy

            if to_rec:
                # t is updated manually here to secure floating point
                # equality check in the buffer updating block
                np.testing.assert_approx_equal(t, next_trec, significant=14)
                t = next_trec
                states_buffer[buffer_pos] = R
                times_buffer[buffer_pos] = t
                buffer_pos += 1
                rec_count += 1
                if dt_record is not None:
                    next_trec = min(dt_record * rec_count, tstop)

                if logging:
                    print(
                        f"{ctime(datetime.now().timestamp())} | {t=} | {it=}",
                        file=sys.stderr,
                    )
            elif nan_timestep or timestep < dt_min:
                if logging:
                    warnings.warn(
                        f"Aborting run with {timestep=} (minvalue is {dt_min=})"
                    )
                states_buffer[buffer_pos] = R
                times_buffer[buffer_pos] = t
                buffer_pos += 1
                rec_count += 1
                if logging:
                    print(
                        f"{ctime(datetime.now().timestamp())} | {t=} | {it=}",
                        file=sys.stderr,
                    )

            if (
                buffer_pos == buffer_size
                or t == tstop
                or nan_timestep
                or timestep < dt_min
            ):
                cls._update_from_buffer(states, states_buffer[:buffer_pos])
                cls._update_from_buffer(times, times_buffer[:buffer_pos])
                progress.update(completed=t, refresh=True)
                buffer_pos = 0

            it += 1

            if nan_timestep:
                metadata["stop_reason"] = "NaN timestep"
                break
            elif timestep < dt_min:
                metadata["stop_reason"] = "reached minimal timestep"
                break

        progress.update(completed=tstop, refresh=True)
        progress.stop()

        return Record(
            times=np.concatenate(times),
            states=np.concatenate(states),
            metadata=metadata,
        )


class RK4(Solver):
    """
    This is the Runge-Kutta method, with order 4
    https://en.wikipedia.org/wiki/Runge–Kutta_methods
    """

    @staticmethod
    def step(
        R: np.ndarray, time: FloatLike, timestep: FloatLike, func: Callable
    ) -> None:
        # - I'm avoiding premature optimisation, so I'm keeping the details of time positioning (t+dt)
        #   even though it should make no difference in my applications
        k1 = func(time, R)
        k2 = func(time + timestep / 2, R + k1 * timestep / 2)
        k3 = func(time + timestep / 2, R + k2 * timestep / 2)
        k4 = func(time + timestep, R + k3 * timestep)
        R += timestep / 6 * (k1 + 2 * k2 + 2 * k3 + k4)



import json
import os
from pathlib import Path
from typing import Any
from typing import Optional

import numpy as np

from dustyn._typing import PathLike


class Record:
    def __init__(
        self,
        times: np.ndarray,
        states: np.ndarray,
        metadata: Optional[dict[str, Any]] = None,
        dirname: Optional[str] = None,
    ):
        self.times = times
        self.states = states
        self.metadata = metadata
        self.dirname = dirname

    def __len__(self):
        return len(self.times)

    @property
    def shape(self):
        return self.states.shape

    def save(
        self, dirname: PathLike, force: bool = False, extra: Optional[dict] = None
    ) -> None:
        dirpath = Path(dirname)
        if not dirpath.is_dir():
            os.makedirs(dirpath)
        elif not force:
            raise FileExistsError(dirname)

        with open(dirpath / "metadata.json", "w") as fh:
            json.dump(self.metadata, fh, indent=2)
            fh.write("\n")

        fields = {"times": self.times, "states": self.states}
        if extra is not None:
            fields.update(extra)
        for name, field in fields.items():
            np.save(dirpath / f"{name}.npy", field)

    @classmethod
    def load(cls, dirname: PathLike, full=False) -> "Record":
        with open(Path(dirname, "metadata.json")) as fh:
            metadata = json.load(fh)

        times = np.load(Path(dirname, "times.npy"))
        states = np.load(Path(dirname, "states.npy"))

        rec = cls(times=times, states=states, metadata=metadata, dirname=str(dirname))
        if full:
            rec.load_extra()
        return rec

    def load_extra(self) -> None:
        if self.dirname is None:
            raise ValueError(
                "This Record instance wasn't loaded from a directory, hence, no extra field is available."
            )
        for file in Path(self.dirname).glob("*.npy"):
            field_name = str(file.stem)
            if hasattr(self, field_name):
                continue
            setattr(self, field_name, np.load(file))



import abc

import numpy as np

from dustyn._typing import FloatLike


class EvolutionModel(abc.ABC):
    @abc.abstractmethod
    def evolve(self, t: float, R: np.ndarray, **kwargs) -> np.ndarray:
        """Compute the time evolution of a state vector R"""
        pass

    @abc.abstractmethod
    def get_max_timestep(
        self,
        R: np.ndarray,
    ) -> FloatLike:
        pass



import abc

import numpy as np

from dustyn._typing import FloatLike
from dustyn.core.medium import Medium
from dustyn.core.newton import Newton
from dustyn.core.newton import NewtonSpherical3D


class Disk(Newton, abc.ABC):
    def __init__(
        self, medium: Medium, short_friction_security: FloatLike = 0.0, **kwargs
    ):
        """
        short_friction_security is disabled by default.
        Use a real number between 0 (excluded) and 1 (included) to activate it.
        """
        if short_friction_security > 1:
            raise ValueError
        self.medium = medium
        self._sf_security = short_friction_security
        self._short_friction = False
        super().__init__(**kwargs)

    @abc.abstractmethod
    def _evolve_short_friction(self, t: FloatLike, R: np.ndarray) -> np.ndarray:
        pass

    @abc.abstractmethod
    def get_stopping_time(self, R: np.ndarray) -> FloatLike:
        pass

    @abc.abstractmethod
    def get_drag(self, R: np.ndarray):
        pass

    def evolve(self, t: float, R: np.ndarray, **kwargs) -> np.ndarray:
        if self._short_friction:
            # because the following algo involves geometric terms
            # it is deleguated to child classes where the geometry is specified
            return self._evolve_short_friction(t, R)
        ret = super().evolve(t, R, **kwargs)
        ret[: self.dimensionality] += self.get_drag(R)
        return ret

    @abc.abstractmethod
    def get_gas_velocity(self, position: np.ndarray) -> np.ndarray:
        pass

    def get_max_timestep(
        self,
        R: np.ndarray,
    ) -> FloatLike:
        dt_kep = super().get_max_timestep(R)
        dt_drag = 2 * self.get_stopping_time(R)

        min_dt = self._sf_security * dt_kep
        self._short_friction = min_dt > dt_drag
        if self._short_friction:
            return min_dt
        return min(dt_kep, dt_drag)


class SphericalDisk3D(Disk, NewtonSpherical3D, abc.ABC):
    def get_gas_velocity(self, position: np.ndarray) -> np.ndarray:
        return np.array(
            [
                self.medium.get(f"velocity_{k}", *position)
                for k in ("r", "theta", "phi")
            ],
            dtype="float64",
        )

    def get_drag(self, R: np.ndarray):
        vp = R[: self.dimensionality]
        vg = self.get_gas_velocity(R[self.dimensionality :])
        delta_v = vg - vp
        return delta_v / self.get_stopping_time(R)

    def _evolve_short_friction(self, t: FloatLike, R: np.ndarray) -> np.ndarray:
        # this method is special in that is will update the state vector R itself,
        # which isn't normally done in other `evolve` methods

        # get acceleration from gravity and geometry
        gacc = NewtonSpherical3D.evolve(self, t, R)[: self.dimensionality]

        ret = np.zeros_like(R)
        # We don't evolve v, we set it by hand (nasty trick, granted)
        # force v to be what we want
        vr, vtheta, vphi = R[: self.dimensionality] = (
            self.get_gas_velocity(R[self.dimensionality :])
            + self.get_stopping_time(R) * gacc
        )

        # update positions following the current velocity
        # this is exactly the same as what is done in NewtonSpherical3D.evolve
        # maybe there's a smart way to avoid code duplication
        r, theta, phi = R[self.dimensionality :]
        ret[self.dimensionality :] = [vr, vtheta / r, vphi / (r * np.sin(theta))]
        return ret


# and now for some concrete implementations
class ConstantTauSphereDisk(SphericalDisk3D):
    def __init__(self, stopping_time: FloatLike = 1.0, **kwargs):
        self.stopping_time = stopping_time
        super().__init__(**kwargs)

    def get_stopping_time(self, R: np.ndarray) -> FloatLike:
        return self.stopping_time


class ConstantStokesSphereDisk(SphericalDisk3D):
    def __init__(self, stokes: FloatLike = 1.0, **kwargs):
        self.stokes = stokes
        super().__init__(**kwargs)

    def get_stopping_time(self, R: np.ndarray) -> FloatLike:
        return (
            self.stokes * self.__class__.spherical_radius_squared(R) ** 0.75 * self.GM
        )



"""
A medium object represents the state of the fluid in which particles evolve.
"""
import abc
from typing import Callable
from typing import Mapping
from typing import Sequence

import numpy as np
import sympy as sp


class Medium(abc.ABC):
    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def get(self, field: str, *args) -> np.ndarray:
        pass


class AnalyticMedium(Medium):
    def __init__(self, symbols: Sequence[sp.Symbol], defs: Mapping[str, sp.Expr]):
        self._lambdas = {
            name: AnalyticMedium.lambdify(symbols, expr) for name, expr in defs.items()
        }

    def get(self, field: str, *args) -> np.ndarray:
        if field not in self._lambdas:
            raise KeyError(
                f"Unknown value `{field=}`. Known fields are {list(self._lambdas.keys())}."
            )
        return self._lambdas[field](*args)

    @staticmethod
    def lambdify(symbols: Sequence[sp.Symbol], expression: sp.Expr) -> Callable:
        """This is specialized replacement for `sympy.lambdify`
        that produce shape-preserving numpy lambdas, with `sympy.lambdify`
        returns scalars when passed constant expressions.

        See https://github.com/sympy/sympy/issues/5642
        """
        f = sp.lambdify(symbols, expression, "numpy")
        if isinstance(expression, (int, float, complex, np.number)):
            return np.vectorize(f)
        return f



import inspect
import weakref
from typing import Callable
from typing import Optional

import numpy as np
from more_itertools import always_iterable
from sympy import pi as SPI

from dustyn.core.medium import Medium


class SphereSlicePlot:
    def __init__(
        self,
        ax,
        fig,
        *,
        grid: np.ndarray,
        medium: Medium,
        background: Optional[np.ndarray] = None,
        zscale="lin",
        zlabel: Optional[str] = None,
        cbar_orientation: str = "horizontal",
        **kwargs,
    ):
        # note that I'm providing a "simple" api to set logscale in the z direction
        # because I could not find a satisfying way to do it with a single argument
        # to mpl's api. It will not work with negative values (no symlog implementation for now)

        if cbar_orientation not in ("horizontal", "vertical"):
            raise ValueError(
                f"Unknown value for `{cbar_orientation=}` (expected 'horizontal' or 'vertical')."
            )
        if zscale not in ("lin", "log"):
            raise ValueError(
                f"Unknown value for `{zscale=}` (expected 'lin' or 'log')."
            )

        # grid is assumed to be written in the canonical order (r, theta)
        # but matplotlib only handles polar projections with
        # "x" as the angular coordinate, so we unpack here and will use those components
        # in a reversed order
        self.rg, self.thetag = grid
        self.fig = weakref.proxy(fig)
        self.ax = weakref.proxy(ax)
        self.medium = weakref.proxy(medium)

        ax.set(xlim=(self.thetag[0, 0], self.thetag[-1, -1]))
        self._setup_curvilinear_axis()

        if background is None:
            return

        if zscale == "log":
            background = np.log10(background)

        im = ax.contourf(self.thetag, self.rg, background, **kwargs)

        cbar_kwargs = {"orientation": cbar_orientation}
        if zlabel is not None:
            cbar_kwargs["label"] = r"$%s$" % zlabel
        cbar = fig.colorbar(im, ax=ax, **cbar_kwargs)
        if cbar_kwargs["orientation"] == "horizontal":
            cbar.ax.tick_params(rotation=45)
            axis = cbar.ax.xaxis
        else:
            axis = cbar.ax.yaxis

        if zscale == "log":
            levels = range(
                int(np.ceil(background.min())), int(np.ceil(background.max()))
            )
            axis.set_ticks(levels)
            axis.set_ticklabels(["$10^{%d}$" % level for level in levels])

    def _setup_curvilinear_axis(self, div: int = 6):
        if self.ax.name == "polar":
            # this is relevant only if the x coordinate is the latitude (theta) in spherical
            self.ax.set(
                rorigin=0,
                theta_zero_location="N",  # theta = 0 is on the vertical axis
                theta_direction=-1,  # set clock-wise counting
            )
            self.ax.grid(False)

        # the x bounds are actually already stored in radian
        # only the xticklabels are written in degrees...
        int_bounds = np.ceil(np.array(self.ax.get_xlim()) / (np.pi / div)).astype(
            "int64"
        )
        ticks_mul = np.arange(int_bounds[0], int_bounds[1])

        if ticks_mul[-1] - ticks_mul[0] >= div * 2:
            # remove the last label if it is redundant to avoid overlaps
            ticks_mul = ticks_mul[:-1]
        self.ax.set_xticks(np.pi / div * ticks_mul)

        labels = [
            r"$%s$" % str(int(m) * SPI / div).replace("pi", r"\pi").replace("*", "")
            for m in ticks_mul
        ]
        self.ax.set_xticklabels(labels)

    def callbacks(self, *callbacks: Callable):
        sig = ["ax", "fig", "r_grid", "theta_grid", "medium"]
        for cb in always_iterable(callbacks):
            params = list(inspect.signature(cb).parameters)
            if params != sig:
                raise TypeError(
                    f"Received callback {cb} with signature {params}, expected {sig}."
                )
            cb(self.ax, self.fig, self.rg, self.thetag, self.medium)



"""
This module defines specialized classes to read self similar 1D simulation
datasets from Lesur 2021, available at https://github.com/glesur/PPDwind
"""
import re
import warnings
from functools import cached_property
from typing import Callable
from typing import Optional

import numpy as np
from scipy.interpolate import interp1d

from dustyn._typing import FloatLike
from dustyn._typing import PathLike
from dustyn.core.medium import Medium
from dustyn.disk import SphericalDisk3D


class SSWDataset:
    HEADER_SIZE = 7

    def __init__(self, fname: PathLike) -> None:
        columns = self.get_colunms(fname)
        self._data = self.load_as_ndarray(fname, columns)
        self._fields = columns

    def __getitem__(self, key: str) -> np.ndarray:
        return self._data[key]

    def get_colunms(self, fname: PathLike) -> list[str]:
        with open(fname) as fh:
            for _ in range(self.HEADER_SIZE - 1):
                next(fh)
            header = fh.readline()
        return re.sub(r"\([\w\s]+\)\s", " ", header).split()

    def load_as_ndarray(
        self, fname: PathLike, columns: Optional[list[str]] = None
    ) -> np.ndarray:
        if columns is None:
            columns = self.get_colunms(fname)
        return np.loadtxt(
            fname,
            skiprows=self.HEADER_SIZE,
            dtype=[(name, "float64") for name in columns],
        )


class SSWMedium(Medium):
    """
    Self-similarity (SS) is central here.
    SS (+ axisymmetry) is equivalent to assuming that any function of space is separable as

    f(r, theta) = f1(r) x f2(theta)
    with
    f1(r) = r^zeta
    where zeta is a constant. Note that, as discussed in the paper, this description
    is valid with either r representing the spherical OR cylindrical radius, though the value
    of zeta and the f2 functions require appropriate transformations to convert between
    coordinate systems.

    In this implementation, f1 is expressed via the `get_zeta` method,
    while functions f2 are stored as 1D anonymous function in `self._lambdas`
    """

    # This is a constant in the whole paper, see section 3.1
    aspect_ratio = 0.1

    self_similar_scaling_exponents: dict[str, float] = {
        # density
        "rho": -1.5,
        # thermal pressure
        "P": -2.5,
        # velocity
        "v": -0.5,
        # magnetic field
        "B": -1.25,
        # ambipolar number
        "Lambda_A": 0.0,
        # Ohmic number
        "Rm": 0.0,
    }

    field_aliases: dict[str, str] = {
        "density": "rho",
        "velocity_r": "vr",
        "velocity_theta": "vtheta",
        "velocity_phi": "vphi",
    }

    def __init__(self, fname: PathLike, **interp_kwargs):
        self.ds = SSWDataset(fname)

        self._lambdas: dict[str, Callable] = {
            field: interp1d(self.ds["theta"], self.ds[field], **interp_kwargs)
            for field in self.ds._fields
        }

        # models are assumed locally isothermal, i.e., sound speed
        # is a pure function of the cylindrical radius R
        # In other words, we need to return the corresponding midplane value, which reads
        # cs_mid = H(R) * Omega_K(R)
        #        = h * R * sqrt(GM/R^3)
        #        = h * r * sin(theta) * sqrt(GM/(r^3 sin^3(theta)))
        #        = r^(-1/2) * h * sqrt(GM/sin(theta))
        # where R is the cylindrical radius here, while r is the spherical radius,
        # and h is the aspect ratio (which is constant in our application)
        # NOTE: I'm assuming the raw data is always using GM=1
        self._lambdas["sound_speed"] = (
            lambda theta: self.aspect_ratio * np.sin(theta) ** -0.5
        )

    @cached_property
    def field_list(self) -> list[str]:
        return list(self._lambdas.keys())

    def get(self, field: str, *args) -> np.ndarray:

        # simulations are actually are axisymetric so there the phi component is not used
        # with PEP 622 we could more easily support flexible inputs as
        # match args:
        #     case r, theta, _phi:
        #         pass
        #     case r, theta:
        #         pass
        #     case _:
        #         raise TypeError("At least two arguments expected.")
        r, theta, phi = args

        field = self.__class__.field_aliases.get(field, field)

        if field not in self._lambdas:
            raise KeyError(
                f"Unknown value `{field=}`. Known fields are {self.field_list}."
            )

        try:
            rv = self._lambdas[field](theta)
        except ValueError:
            msg = (
                "Invalid values encountered. "
                "Consider passing `bounds_error=False` at instance creation. "
            )
            if (min_eff := theta.min()) < (min_valid := self.ds["theta"].min()):
                msg += f"Min possible theta is {min_valid} while input has min(theta) = {min_eff} "
            if (max_eff := theta.max()) > (max_valid := self.ds["theta"].max()):
                msg += f"Max possible theta is {max_valid} while input has max(theta) = {max_eff}"
            warnings.warn(msg)
            if hasattr(r, "shape"):
                rv = np.full_like(r, np.nan)
            else:
                # assume scalar input
                rv = np.nan

        return r ** self._get_zeta(field) * rv

    def _get_zeta(self, field: str) -> float:
        """Retrieve the appropriate scaling factor"""
        if field == "sound_speed":
            return self.self_similar_scaling_exponents["v"]
        for prefix, zeta in self.self_similar_scaling_exponents.items():
            if field.startswith(prefix):
                return zeta
        raise KeyError(f"Could not find appropriate scaling factor for {field=}")


class SSWDisk(SphericalDisk3D):
    def __init__(self, mass_to_surface: FloatLike, **kwargs):
        # I define the mass to surface ratio as the product of particle radius
        # and particle intrinsic density, which, akin to "GM" in equations of
        # motion for test particles, only ever appear as a product.
        self.mass_to_surface = mass_to_surface
        super().__init__(**kwargs)

    def get_stopping_time(self, R: np.ndarray) -> FloatLike:
        # This assumes the Epstein regime is relevant
        # TODO: check implementation (a dimensionless geometric factor may be missing)
        return self.mass_to_surface / (
            self.medium.get("density", *R[3:]) * self.medium.get("sound_speed", *R[3:])
        )



import abc
import warnings
from typing import Union

import numpy as np

from .algebra import cross_product
from .evolve import EvolutionModel
from .transform import spherical2cartesian
from dustyn._typing import FloatLike


class Newton(EvolutionModel, abc.ABC):
    dimensionality: int
    geometry: str

    def __init__(self, GM: FloatLike = 1.0):
        self.GM = GM

    @abc.abstractmethod
    def hamiltonian(self, R: np.ndarray) -> Union[FloatLike, np.ndarray]:
        """Compute the hamiltonian of a given state.
        This quantity is supposed to be conserved through the integration as is
        useful for testing.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def spherical_radius_squared(cls, R: np.ndarray) -> FloatLike:
        pass

    def get_max_timestep(
        self,
        R: np.ndarray,
    ) -> FloatLike:
        # the relevant maximal timestep dictated by the gravitational force reads
        # dt = 1 / Omega_K
        #    = sqrt(r^3 / GM)
        #    = ((r^2)^3/2)^1/2 * (GM)^-1/2
        #    = r2^3/4 * (GM)^-1/2
        return self.__class__.spherical_radius_squared(R) ** 0.75 * self.GM ** (-0.5)


class NewtonCartesian(Newton, abc.ABC):
    geometry = "cartesian"

    @classmethod
    def spherical_radius_squared(cls, R: np.ndarray) -> FloatLike:
        return np.sum(R[cls.dimensionality :] ** 2, axis=0)


class NewtonCartesian2D(NewtonCartesian):
    dimensionality = 2

    def evolve(self, t: float, R: np.ndarray, **kwargs) -> np.ndarray:
        """Compute the time evolution of a state vector R
        such that R = [vx, vy, x, y]
        """
        vx, vy, x, y = R
        phi = np.arctan2(y, x)
        r2 = self.__class__.spherical_radius_squared(R)

        tot_force = -self.GM / r2
        return [tot_force * np.cos(phi), tot_force * np.sin(phi), vx, vy]

    def hamiltonian(self, R: np.ndarray) -> Union[float, np.ndarray]:
        vx, vy, x, y = R
        r2 = self.__class__.spherical_radius_squared(R)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", r"divide by zero encountered in true_divide"
            )
            return 0.5 * (vx ** 2 + vy ** 2) - self.GM / np.sqrt(r2)


class NewtonCartesian3D(NewtonCartesian):
    dimensionality = 3

    def evolve(self, t: float, R: np.ndarray, **kwargs) -> np.ndarray:
        """Compute the time evolution of a state vector R
        such that R = [vx, vy, vz, x, y, z]
        """
        vx, vy, vz, x, y, z = R
        phi = np.arctan2(y, x)

        rho = np.sqrt(x ** 2 + y ** 2)  # cylindrical radius
        theta = np.arctan2(rho, z)

        r2 = self.__class__.spherical_radius_squared(R)

        tot_force = -self.GM / r2
        return [
            tot_force * np.cos(phi) * np.sin(theta),
            tot_force * np.sin(phi) * np.sin(theta),
            tot_force * np.cos(theta),
            vx,
            vy,
            vz,
        ]

    def hamiltonian(self, R: np.ndarray) -> Union[float, np.ndarray]:
        vx, vy, vz, x, y, z = R

        r2 = self.__class__.spherical_radius_squared(R)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", r"divide by zero encountered in true_divide"
            )
            return 0.5 * (vx ** 2 + vy ** 2 + vz ** 2) - self.GM / np.sqrt(r2)


class NewtonCurvilinear(Newton, abc.ABC):
    # regroup polar, spherical and cylindrical geometries
    @classmethod
    def spherical_radius_squared(cls, R: np.ndarray) -> FloatLike:
        return R[cls.dimensionality] ** 2


class NewtonPolar2D(NewtonCurvilinear):
    """this is an exercise before getting to full 3D spherical
    The coordinate system is de-facto polar (r, phi).
    """

    dimensionality = 2
    geometry = "cylindrical"

    def evolve(self, t: float, R: np.ndarray, **kwargs) -> np.ndarray:
        """Compute the time evolution of a state vector R
        such that R = [vr, vphi, r, phi]
        """
        vr, vphi, r, phi = R

        g = -self.GM / r ** 2
        return [
            g + vphi ** 2 / r,
            -vr * vphi / r,
            vr,
            vphi / r,
        ]

    def hamiltonian(self, R: np.ndarray) -> Union[float, np.ndarray]:
        vr, vphi, r, phi = R

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", r"divide by zero encountered in true_divide"
            )
            return 0.5 * (vr ** 2 + vphi ** 2) - self.GM / r


class NewtonSpherical3D(NewtonCurvilinear):
    dimensionality = 3
    geometry = "spherical"

    def evolve(self, t: float, R: np.ndarray, **kwargs) -> np.ndarray:
        """Compute the time evolution of a state vector R
        such that R = [vr, vtheta, vphi, r, theta, phi]
        """
        vr, vtheta, vphi, r, theta, phi = R

        g = -self.GM / r ** 2
        return np.array(
            [
                g + (vphi ** 2 + vtheta ** 2) / r,
                (vphi ** 2 / np.tan(theta) - vr * vtheta) / r,
                -vphi * (vr + vtheta / np.tan(theta)) / r,
                vr,
                vtheta / r,
                vphi / (r * np.sin(theta)),
            ]
        )

    def hamiltonian(self, R: np.ndarray) -> Union[float, np.ndarray]:
        vr, vtheta, vphi, r, theta, phi = R

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", r"divide by zero encountered in true_divide"
            )
            return 0.5 * (vr ** 2 + vphi ** 2 + vtheta ** 2) - self.GM / r

    def angular_momentum(self, R: np.ndarray) -> np.ndarray:
        """Compute the CARTESIAN components of the angular momentum."""
        if R.ndim > 2:
            raise NotImplementedError(
                "This may work but I haven't tried it yet. "
                "Remove this `raise` statement if you have faith."
            )

        ncols = 1 if R.ndim == 1 else R.shape[1]
        r = np.zeros((3, ncols), dtype="float64").squeeze()
        r[0] = R[3]
        v = np.array(R[:3])
        L = cross_product(r, v)
        return spherical2cartesian(position=R[3:7], vector=L)



import os
from typing import Iterable
from typing import TypeVar
from typing import Union

import numpy as np

# PEP 484 specify that integers are acceptable where floats are required
FloatLike = Union[float, np.floating]

# types that are supposed to support pathlib.Path (natively, str and classes implementing the __fspath__ protocol)
PathLike = Union[str, os.PathLike[str]]

T = TypeVar("T")

SingleOrDouble = Union[T, tuple[T, T]]
SingleOrMultiple = Union[T, Iterable[T]]
